/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.bridge;

import java.io.File;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;

import javax.sql.DataSource;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.extractor.Extractor;
import org.xquark.extractor.ExtractorConnection;
import org.xquark.schema.SchemaManager;
import org.xquark.xml.xdbc.XMLDBCException;

/**
 * Objects of this class encapsulate a JDBC connection and provide access to both
 * the XQuery and insertion capabilities of the XQBridge.<BR>
 * Objects of this class are thread-safe and can be reused multiple times.
 * However, as they encapsulate a JDBC connection, it is recommended to use
 * different objects with different JDBC connections for different threads.
 */
public class XQBridge {
	private static final String RCSRevision = "$Revision: 1.19 $";
	private static final String RCSName = "$Name:  $";

	// Initialisation parameters (initialisation is delayed, so that the class loader can be specified)
	private Connection jdbcConnection;
	private DataSource dataSource;
	private String configURI;
	private File configFile;
	private String driver;
	private String jdbcURL;
	private String userName;
	private String password;

	private Extractor extractor;
	private HashMap mappingMetadata = new HashMap();
	private SchemaManager schemaManager = null;
	private PrintStream log = null;
	private ClassLoader classLoader = null;

	/** The constructor for the XQBridge object. This constructor allocates its own JDBC connection.
	 * The close() method must be called on the object to free the JDBC resources.
	 * @param confURI the XQBridge configuration file URI
	 */
	public XQBridge(String confURI) {
		if (confURI == null)
			throw new NullPointerException("Illegal null value for configuration file URI");
		configURI = confURI;
	}

	public XQBridge(File confFile) {
		if (confFile == null)
			throw new NullPointerException("Illegal null value for configuration file");
		configFile = confFile;
	}

	/**
	 * The constructor for the XQBridge object.
	 * @param dataSource the datasource used to access the database.
	 */
	public XQBridge(DataSource dataSource) {
		if (dataSource == null)
			throw new NullPointerException("Illegal null value for data source");
		this.dataSource = dataSource;
	}

	/**
	 * The constructor for the XQBridge object. The specified datasource is used to access the database,
	 * and overrides connection information provided in the configuration file.
	 * @param dataSource the datasource used to access the database.
	 * @param confFile the XQBridge configuration file
	 */
	public XQBridge(DataSource dataSource, File confFile) {
		if (dataSource == null)
			throw new NullPointerException("Illegal null value for data source");
		if (confFile == null)
			throw new NullPointerException("Illegal null value for configuration file");
		this.dataSource = dataSource;
		this.configFile = confFile;
	}

	/**
	 * The constructor for the XQBridge object. 
	 * @param driver the JDBC driver
	 * @param url the JDBC URL
	 * @param userName the user name
	 * @param password the password
	 */
	public XQBridge(String driver, String url, String userName, String password) {
		if (url == null)
			throw new NullPointerException("Illegal null value for URL");
		if (userName == null)
			throw new NullPointerException("Illegal null value for user name");
		if (password == null)
			throw new NullPointerException("Illegal null value for user password");
		this.driver = driver;
		this.jdbcURL = url;
		this.userName = userName;
		this.password = password;
	}

	/** The constructor for the XQBridge object. The specified JDBC connection is used to access the database.
	 * This connection should not be used while the XQBridge object is active. 
	 * Use XQBridge.getConnection() to retrieve a usable JDBC connection.
	 * @param jdbcConn the JDBC connection used to access the database.
	 * @deprecated
	 */
	public XQBridge(Connection jdbcConn) {
		if (jdbcConn == null)
			throw new NullPointerException("Illegal null value for connection");
		jdbcConnection = jdbcConn;
	}

	/** The constructor for the XQBridge object. The specified JDBC connection is used to access the database,
	 * and overrides connection information provided in the configuration file. This connection should not be used
	 * while the XQBridge object is active. Use XQBridge.getConnection() to retrieve a usable JDBC connection.
	 * @param jdbcConn the JDBC connection used to access the database.
	 * @param confFile the XQBridge configuration file
	 * @deprecated
	 */
	public XQBridge(Connection jdbcConn, String confURI) {
		if (jdbcConn == null)
			throw new NullPointerException("Illegal null value for connection");
		if (confURI == null)
			throw new NullPointerException("Illegal null value for configuration file URI");
		jdbcConnection = jdbcConn;
		configURI = confURI;
	}

	private Extractor getExtractor() throws XMLDBCException {
		if (extractor == null) {
			if (configURI == null && configFile != null) {
				try {
					configURI = configFile.toURL().toString();
				} catch (MalformedURLException ex) {
					throw new XMLDBCException("Invalid configuration file name " + configFile, ex);
				}
			}
			if (dataSource != null) {
				extractor = new Extractor(dataSource, configURI);
			} else if (jdbcConnection != null) {
				extractor = new Extractor(jdbcConnection, configURI);
			} else if (jdbcURL != null) {
				extractor = new Extractor(jdbcURL, userName, password);
			} else {
				extractor = new Extractor(configURI, getClassLoader());
			}
		}
		return extractor;
	}
	/**
	 * Returns a JDBC connection used to access the underlying database. This
	 * connection must be closed after use.
	 * 
	 * @return a JDBC connection. Null when no connection to the database is
	 * available.
	 * @deprecated This method is linked to the deprecated constructor
	 * {@link #XQBridge(Connection, String)}.
	 */
	public Connection getConnection() throws XMLDBCException {
		Connection conn = null;
		try {
			conn = getExtractor().getJdbcDataSource().getConnection();
		} catch (SQLException ex) {
			throw new XMLDBCException("Could not create connection", ex);
		}
		return conn;
	}

	/**
	 * Returns an extractor XML/DBC connection that can be used to execute
	 * XQuery statements over the database.
	 * @return an XMLConnection object
	 */
	public ExtractorConnection getXMLConnection() throws XMLDBCException {
		return getExtractor().getExtractorConnection();
	}

	public void close() throws XMLDBCException {
		if (extractor != null)
			extractor.close();
	}

	/**
	 * Set the class loader used for loading drivers, mappings and user generators.
	 * @param loader ClassLoader 
	 */
	public void setClassLoader(ClassLoader loader) {
		classLoader = loader;
	}

	/**
	 * Returns the class loader used for loading drivers, mappings and user generators.
	 */
	public ClassLoader getClassLoader() {
		return classLoader != null ? classLoader : this.getClass().getClassLoader();
	}

	/**
	 * Creates a mapping object containing the processed version of the specified mapping file.
	 * This object can subsequently be used to store XML documents in the database, according
	 * to the specified mapping.<BR>
	 * The mapping file is specified as a SAX input source. If the systemId property of the input source is set,
	 * the object will first be looked up, based on this property, in the mapping cache, and only be created if
	 * not found.
	 * @deprecated Prefer the string methods because they do not create a problem
	 * for resolving relative URIs in mapping file schemaLocation attribute.
	 * @param source a SAX input source encapsulating the mapping file.
	 * @return a Mapping object
	 * @throws XMLDBCException if an error occured while loading the mapping file. The underlying exception provides the
	 * root cause for the exception
	 */
	public Mapping getMapping(InputSource source) throws XMLDBCException {
		return loadMapping(source, false);
	}

	/**
	 * Creates a mapping object containing the processed version of the specified mapping file.
	 * This object can subsequently be used to store XML documents in the database, according
	 * to the specified mapping.<BR>
	 * The mapping file is specified as a SAX input source. If the systemId property of the input source is set,
	 * and the reload flag is false, the object will first be looked up, based on this property,
	 * in the mapping cache, and only be created if not found.
	 * @deprecated Prefer the string methods because they do not create a problem
	 * for resolving relative URIs in mapping file schemaLocation attribute.
	 * @param source a SAX input source encapsulating the mapping file.
	 * @param reload true if the mapping file must be reloaded even if found in the cache.
	 * @throws XMLDBCException if an error occured while loading the mapping file. The underlying exception provides the
	 * root cause for the exception
	 * @return a Mapping object
	 */
	public Mapping getMapping(InputSource source, boolean reload) throws XMLDBCException {
		return loadMapping(source, reload);
	}

	/**
	 * Creates a mapping object containing the processed version of the specified mapping file.
	 * This object can subsequently be used to store XML documents in the database, according
	 * to the specified mapping.<BR>
	 * the object will first be looked up, based on the file URI, in the mapping cache,
	 * and only be created if not found.
	 * @param fileURI the URI of the mapping file.
	 * @return a Mapping object
	 * @throws XMLDBCException if an error occured while loading the mapping file. The underlying exception provides the
	 * root cause for the exception
	 */
	public Mapping getMapping(String fileURI) throws XMLDBCException {
		return loadMapping(new InputSource(fileURI), false);
	}

	/**
	 * Creates a mapping object containing the processed version of the specified mapping file.
	 * This object can subsequently be used to store XML documents in the database, according
	 * to the specified mapping.<BR>
	 * If the reload flag is false, the object will first be looked up, based on the file URI,
	 * in the mapping cache, and only be created if not found.
	 * @param fileURI the URI of the mapping file.
	 * @param reload true if the mapping file must be reloaded even if found in the cache.
	 * @return a Mapping object
	 * @throws XMLDBCException if an error occured while loading the mapping file. The underlying exception provides the
	 * root cause for the exception
	 */
	public Mapping getMapping(String fileURI, boolean reload) throws XMLDBCException {
		return loadMapping(new InputSource(fileURI), reload);
	}

	private Mapping loadMapping(InputSource source, boolean reload) throws XMLDBCException {
		return new Mapping(getExtractor().getJdbcDataSource(), source, reload, getClassLoader(), getSchemaManager(), mappingMetadata);
	}

	private SchemaManager getSchemaManager() {
		if (schemaManager == null)
			resetSchemas();
		return schemaManager;
	}

	private void resetSchemas() {
		schemaManager = new SchemaManager();
		String resource = getClass().getResource("/org/xquark/bridge/resources/mapping.xsd").toString();

		try {
			schemaManager.loadSchema(new InputSource(resource));
		} catch (SAXException e) {
			// Should not occur
		}
	}

	/**
	 * Loads an XML schema in memory. This is useful for bypassing the automatic schema loading based
	 * upon the schemaLocation attribute (e.g. if there is no such attribute).
	 */
	public void loadSchema(InputSource source) throws SAXException {
		getSchemaManager().loadSchema(source);
	}

	/**
	 * Empties the mappings cache. All Mapping objects that were not closed will continue
	 * working with the old metadata. 
	 * @param resetSchemas if the schema cache must be cleared as well
	 */
	public void resetMappings(boolean resetSchemas) {
		mappingMetadata.clear();
		if (resetSchemas)
			resetSchemas();
	}

}
